
#pragma once

// Genode includes
#include <base/env.h>
#include <base/heap.h>

// protobuf includes
#include <google/protobuf/message.h>
#include <google/protobuf/util/time_util.h>

// gapfruit includes
#include <tee_session/connection.h>
#include <tee/packet_allocator.h>
#include <tee_protobuf/helpers.h>

// stdcxx includes
#include <functional>


namespace Tee {

	using Genode::Allocator_avl;
	using Genode::Env;
	using Genode::Heap;
	using Genode::List;
	using Genode::Signal_context_capability;
	using Genode::error;
	using Genode::memset;

	using Packet_alloc_failed = Tee::Session::Tx::Source::Packet_alloc_failed;

	using google::protobuf::MessageLite;
	using google::protobuf::Message;

	using gapfruit::protobuf::tee::CounterResponse;
	using gapfruit::protobuf::tee::ExecuteResponse;
	using gapfruit::protobuf::tee::QuoteResponse;
	using gapfruit::protobuf::tee::SignedResponse;

	class Connection_asynchronous;

	struct Proto_callbacks;
}


struct Tee::Proto_callbacks
{
	std::function<void(Packet_content_ro)>  message_received;
	std::function<void()>                   counter_request;
	std::function<void()>                   execute_init_request;
	std::function<void()>                   execute_request;
	std::function<void()>                   quote_request;
};


class Tee::Connection_asynchronous
{
	private:

		using Signal_handler = Genode::Signal_handler<Connection_asynchronous>;

		Env&                     _env;
		Heap                     _heap             { _env.ram(), _env.rm() };
		Allocator_avl            _tx_block_alloc   { &_heap };
		Tee::Connection          _connection;
		const char*              _session_label;
		Request_info&            _request_info;
		Execute_request_data&    _execute_request_data;
		Execute_response_data&   _execute_response_data;
		const Proto_callbacks&   _callbacks;

		// signal handlers for connection
		Signal_handler          _rx_data          { _env.ep(), *this, &Connection_asynchronous::_handle_rx };
		Signal_handler          _tx_data          { _env.ep(), *this, &Connection_asynchronous::_handle_tx };

		Signal_context_capability  _counter_request  { };

		// methods used by the signal handlers
		void _handle_rx();
		void _handle_tx();

	public:

		enum { DEFAULT_BUF_SIZE = Tee::Packet_allocator::DEFAULT_PACKET_SIZE * 128 };

		Connection_asynchronous(Genode::Env& env,
		                        const char* label,
		                        const Proto_callbacks& callbacks,
		                        Request_info& request_info,
		                        Execute_request_data& exec_request_data,
		                        Execute_response_data& exec_response_data,
		                        size_t tx_buf_size = DEFAULT_BUF_SIZE,
		                        size_t rx_buf_size = DEFAULT_BUF_SIZE) :
		    _env                   { env },
		    _connection            { env, &_tx_block_alloc, tx_buf_size, rx_buf_size, label },
		    _session_label         { label },
		    _request_info          { request_info },
		    _execute_request_data  { exec_request_data },
		    _execute_response_data { exec_response_data },
		    _callbacks             { callbacks }
		{
			_connection.rx_channel()->sigh_packet_avail(_rx_data);
			_connection.rx_channel()->sigh_ready_to_ack(_rx_data);
			_connection.tx_channel()->sigh_ready_to_submit(_tx_data);
			_connection.tx_channel()->sigh_ack_avail(_tx_data);
		}

		~Connection_asynchronous() = default;

		/**
		 * Sends a protobuf message
		 *
		 * \param msg  The protobuf message to send.
		 * \param sid  The session id. Sinks need to reply using the received
		 *             session id so the source can track sessions.
		 *
		 * \throw Transmission_error  E.g. when there is no sink to send to.
		 * \throw Serialize_failed    Protobuf message could not be serialized.
		 * \throw Packet_alloc_failed Could not allocate packet on connection.
		 */
		void send(const MessageLite& msg, const Tee::Session_id sid);

		/**
		 * Send a CounterResponse
		 *
		 * \param  data          struct containing the data for the response
		 * \param  request_info  struct with the data from the request
		 *
		 * \throw Packet_alloc_failed  Could not allocate packet on connection.
		 * \throw Serialize_failed     Protobuf message could not be serialized.
		 * \throw Transmission_error   E.g. when there is no sink to send to.
		 */
		void send_counter_response(const Counter_response_data& data, const Request_info& request_info);

		/**
		 * Send a ExecuteInitResponse
		 *
		 * \param  data          struct containing the data for the response
		 * \param  request_info  struct with the data from the request
		 *
		 * \throw Packet_alloc_failed  Could not allocate packet on connection.
		 * \throw Serialize_failed     Protobuf message could not be serialized.
		 * \throw Transmission_error   E.g. when there is no sink to send to.
		 */
		void send_execute_init_response(const Execute_init_response_data& data, const Request_info& request_info);

		/**
		 * Send a ExecuteResponse
		 *
		 * \param  data          struct containing the data for the response
		 * \param  request_info  struct with the data from the request
		 *
		 * \throw Packet_alloc_failed  Could not allocate packet on connection.
		 * \throw Serialize_failed     Protobuf message could not be serialized.
		 * \throw Transmission_error   E.g. when there is no sink to send to.
		 */
		void send_execute_response(Execute_response_data& data, const Request_info& request_info);

		/**
		 * Send a CounterResponse
		 *
		 * \param  data          struct containing the data for the response
		 * \param  request_info  struct with the data from the request
		 *
		 * \throw Packet_alloc_failed  Could not allocate packet on connection.
		 * \throw Serialize_failed     Protobuf message could not be serialized.
		 * \throw Transmission_error   E.g. when there is no sink to send to.
		 */
		void send_quote_response(const Quote_response_data& data, const Request_info& request_info);
};
